
#include <stdio.h>
#include <errno.h>

#ifdef WIN32
#include <Winsock2.h>
#else
#include <sys/types.h>
#include <sys/socket.h> 
#include <arpa/inet.h>
#include <fcntl.h>
#define closesocket close
#endif

#include <string.h>
#include <signal.h>

#include <SDL.h>


#define PORT 44455
#define MAXCONN 10
#define BUFSIZE 1024

#define TYPE_DEFAULT 1

typedef struct client {
	int fd;
	char buf[BUFSIZE];
	int buf_end;
} client_t;

client_t clients[MAXCONN];

SDL_Haptic* dev;
SDL_HapticEffect effect;
int effectid;

int serverfd;

int rumble_init() {
	int i;

	if (SDL_Init(SDL_INIT_HAPTIC)) {
		fprintf(stderr, "SDL_Init failed");
		return 1;
	}
	
	fprintf(stderr, "num haptics %i\n", SDL_NumHaptics());

	for (i = 0; i < SDL_NumHaptics(); i++) {
	  if (dev = SDL_HapticOpen(i)) {
	    if (SDL_HapticQuery(dev) & SDL_HAPTIC_SINE) {
	      effect.type = SDL_HAPTIC_SINE;
	      effect.periodic.length = 100;
	      effect.periodic.period = 100;
	      effect.periodic.magnitude = 32767;	      
	      
	      effectid = SDL_HapticNewEffect(dev, &effect);
	    }
	    
	    if (effectid < 0 && SDL_HapticQuery(dev) & SDL_HAPTIC_LEFTRIGHT) {
	      effect.type = SDL_HAPTIC_LEFTRIGHT;
	      effect.leftright.length = 100;
	      effect.leftright.large_magnitude = 32767;
	      effect.leftright.small_magnitude = 32767;   
	      
	      effectid = SDL_HapticNewEffect(dev, &effect);
	    }
	    
	    if (effectid >= 0) {
	      fprintf(stderr, "found usable device: %i\n", i);
	      return 0;
	    }
	  }
	
	  fprintf(stderr, "haptics not usable: %i - %s\n", i, SDL_GetError());
	  SDL_HapticClose(dev);
	}

	SDL_Quit();

	return 1;
}

void rumble_play(int type, int strength, int duration) {
	if (strength < 1 || strength > 100) {
		fprintf(stderr, "rumble_play bad strength: %i\n", strength);
	}

	if (duration < 1 || duration > 10000) {
		fprintf(stderr, "rumble_play bad duration: %i\n", duration);
	}
	
	if (!dev) {
	  return;
	}

	if (effect.type == SDL_HAPTIC_SINE) {
	  effect.periodic.period = ((100.0f / strength) * 100);
	  effect.periodic.length = duration;
	  effect.periodic.magnitude = (Sint16)((strength / 100.0f) * 32767);
	} else {
	  effect.leftright.length = duration;
	  effect.leftright.large_magnitude = (strength / 100.0f) * 32767.0f;	
	  effect.leftright.small_magnitude = (strength / 100.0f) * 32767.0f;	
	}
	SDL_HapticUpdateEffect(dev, effectid, &effect);
	SDL_HapticRunEffect(dev, effectid, 1);
}

void rumble_quit() {
	SDL_Quit();
}

int get_param(char **buf) {
	char param[10];
	char *buf_ptr = *buf;
	char *end_ptr;
	int ret;
	int i;

	for (i = 0; i < 9; i++) {
		char chr = *buf_ptr;

		buf_ptr++;
		if ((chr == ' ') || (chr == '\0')) {
			break;
		}

		param[i] = chr;
	}

	param[i] = '\0';
	*buf = buf_ptr;

	errno = 0;
	ret = (int)strtol(param, &end_ptr, 10);

	if (errno != 0) {
		perror("get_param");
		ret = -1;
	} 
	else if (param == end_ptr) {
		fprintf(stderr, "get_param: not a number\n");
		ret = -1;
	}
  
	return ret;
}

int have_linebreak(char *buf, int len) {
	int i;

	for (i = 0; i < len; i++) {
		if (buf[i] == '\n')
			return i;
	}

	return -1;
}

int consume_buf(char* buf, int len) {
	int cmd_len = 0;
	char *buf_ptr;

	char *param_ptr;
	
	int strength;
	int duration;
	
	buf_ptr = buf;

	while ((cmd_len = have_linebreak(buf_ptr, len)) > -1) {
		*(buf_ptr + cmd_len) = '\0';
		fprintf(stderr, "cmd: %s\n", buf_ptr);

		param_ptr = buf_ptr;

		switch(get_param(&param_ptr)) {
		case TYPE_DEFAULT:

			strength = get_param(&param_ptr);
			duration = get_param(&param_ptr);
			rumble_play(TYPE_DEFAULT, strength, duration);
		}


		len-=(cmd_len + 1);
		buf_ptr+=(cmd_len + 1);
	}

	memmove(buf, buf_ptr, len);

	return len;
}


void clean_exit() {
  int i;
  
  printf("exiting\n");

  rumble_quit();

  for (i = 0; i < MAXCONN; i++) {
    if (clients[i].fd >= 0) {
      closesocket(clients[i].fd);
      clients[i].buf_end = 0;
    }
  }
  close(serverfd);

#ifdef WIN32
  WSACleanup();
#endif
  
  exit(0);
}

void sigint_handler(int arg) {
  clean_exit();
}

int set_nonblocking(int sock) {
#ifdef WIN32
	u_long mode=1;
	return ioctlsocket(sock, FIONBIO, &mode); 
#else
	return fcntl(sock, F_SETFL, O_NONBLOCK);
#endif
}


int main(int argc, char **argv) {

	int i;

	int clientfd;
	fd_set readfds;
	int maxfd;

	struct sockaddr_in addr;

	printf("rumbleserver starting...\n");

	signal(SIGINT, sigint_handler);

	if (rumble_init()) {
		fprintf(stderr, "rumble_init failed\n");
	}
	
	for (i = 0; i < MAXCONN; i++) {
		clients[i].fd = -1;
	}

#ifdef WIN32
	WSADATA wsaData;
	WSAStartup(MAKEWORD(2, 2), &wsaData);
#endif

	if ((serverfd = socket(PF_INET, SOCK_STREAM, 0)) == -1) {
		perror("server socket");
		clean_exit();
	}

	set_nonblocking(serverfd);

	addr.sin_family = AF_INET;
	addr.sin_port = htons(PORT);
	addr.sin_addr.s_addr = inet_addr("127.0.0.1");
	if (addr.sin_addr.s_addr == -1) {
 		fprintf(stderr, "inet_addr error\n");
		clean_exit();
	}
	
	memset(addr.sin_zero, '\0', sizeof(addr.sin_zero));

	if (bind(serverfd, (struct sockaddr *)&addr, sizeof(addr)) == -1) {
		perror("server bind");
		clean_exit();
	}
	
	if (listen(serverfd, 5) == -1) {
		perror("server listen");
		clean_exit();
	}

	while (1) {
		FD_ZERO(&readfds);
		FD_SET(serverfd, &readfds);

		for (i = 0; i < MAXCONN; i++) {
			if (clients[i].fd >= 0) {
				FD_SET(clients[i].fd, &readfds);
			}
		}


		if (serverfd > maxfd)
			maxfd = serverfd;

		select(maxfd + 1, &readfds, NULL, NULL, NULL);
		
		if (FD_ISSET(serverfd, &readfds) &&
		    (clientfd = accept(serverfd, NULL, NULL)) > -1) {
			set_nonblocking(clientfd);
			for (i = 0; i < MAXCONN; i++) {
				if (clients[i].fd < 0) {
					if (clientfd > maxfd)
						maxfd = clientfd;

					clients[i].fd = clientfd;
					fprintf(stderr, "added new client %i\n", clientfd);
					break;
				}
			}

			if (i >= MAXCONN) {
				closesocket(clientfd);
			}
		}
		
		for (i = 0; i < MAXCONN; i++) {
			clientfd = clients[i].fd;

			if (clientfd >= 0 && FD_ISSET(clientfd, &readfds)) {
				int buf_end = clients[i].buf_end;
				int size = recv(clientfd, clients[i].buf + buf_end, BUFSIZE - buf_end, 0);
				
				if (size == 0) {
					fprintf(stderr, "connection closed: %i\n", clientfd);
					closesocket(clientfd);
					clients[i].fd = -1;
					clients[i].buf_end = 0;
				} else if (size > 0) {
					buf_end = consume_buf(clients[i].buf, buf_end + size);
					if (buf_end >= BUFSIZE) {
						fprintf(stderr, "no command in receive buffer. discarding: %i\n", clientfd);
						buf_end = 0;
					}
					clients[i].buf_end = buf_end;
				} else {
					fprintf(stderr, "no data.. this shouldn't happen: %i %i\n", clientfd);
				}
			}
		}
	}

	clean_exit();
	return 0;
}
